/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "alibi.h"

/* These are globals declared in alibi.h */
int insocket = -1, inport_number, nfds=0, nuke=0;
long netccount=0, pingrqcount=0, mrt=0, noncracc=0, noncrdel=0,
  pingrscount=0, lagcount=0, querycount=0, responsecount=0, anscount=0,
  qacc=0, qdel=0, cachedel=0, messbufsize;
time_t s_time, last_relink=0;
fd_set fdset, wrset, tfdset, twrset;
unsigned long serialnum=0, c_hops, c_links, c_cache;
char myhostname[namlen], *messbuf;
struct conlist *cons = (struct conlist *)0;
struct qrtable *qroutes = NULL;
long freespace = cachesize, myaddr, neighborlag=0, unrealtime=0;

#ifdef sig_debug
#include "leak.c"
#endif

/* Read FQDN if it exists.  This file can be used to override the host name
 * and domain claimed by the system if it's wrong.
 */
int
read_FQDN ()
{
  FILE *fp;
  if (!(fp = fopen ("FQDN", "rt")))
    return 0;
  fscanf (fp, "%s", myhostname);
  fclose (fp);
  return 1;
}

void
set_myhostname ()
{
  struct hostent *hp;
  if (!(read_FQDN ()))
  {
    if (gethostname (myhostname, namlen) == -1)
      panic ("gethostname", 1);

    /* Replace myhostname with FQDN -- it's important!!! */
    if (strchr (myhostname, (int)'.') == NULL)
    {
      /* strcat (myhostname, "."); */
      getdomainname (&(myhostname[strlen(myhostname)]), namlen -
        strlen (myhostname));
      myhostname[namlen-1] = '\0';
    }
  }
  strdowncase (myhostname);
  /* Host name must be the canonical FQDN. */
  printf ("I am %s.\n", myhostname);
  fflush (stdout);
  hp = (struct hostent *) gethostbyname (myhostname);
  if (hp == 0)
    panic ("gethostbyname failed on own FQDN!", 1);
  /* Note that gethostid tends to be platform-dependent. */
  myaddr = (long) ntohl (**(unsigned long **)(hp->h_addr_list));
  printf ("My ID number is %ld and my PID is %d.\n", myaddr, getpid ());
}

/* Just translate values of contype into their ASCII meanings. */
char *
title (contype a)
{
  switch (a) {
  case generic:
    return "generic";
  case delete:
    return "delete";
  case connecting:
    return "connecting";
  case new:
    return "new";
  case bozo:
    return "bozo";
  case command:
    return "command";
  case wannabe:
    return "wannabe";
  case near_daemon:
    return "near_daemon";
  case daemon:
    return "daemon";
  case daemon_oob:
    return "daemon_oob";
  case client:
    return "client";
  case resource:
    return "resource";
  default:
    return "???";
  }
}

/* Change handler to the right function (called by log_promo) */
void
rehandle (struct conlist *t)
{
  switch (t->type) {
  case connecting:
    t->handler = connectinghandler; break;
  case new:
    t->handler = newhandler; break;
  case command:
    t->handler = comhandler; break;
  case wannabe:
    t->handler = wannabehandler; break;
  case near_daemon:
    t->handler = nearhandler; break;
  case daemon:
    t->handler = daemonhandler; break;
  case client:
    t->handler = clienthandler; break;
  case resource:
    t->handler = resourcehandler; break;
  default:
    t->handler = bitbucket;
  }
}

/* The /dev/null of message handlers.  Connections being deleted use this. */
void
bitbucket (char *buf, struct conlist *t, struct conlist **cons)
{
  ;
}

void
connectinghandler (char *buf, struct conlist *t, struct conlist **cons)
{
  panic ("Message received on unconnected socket!!!", 1);
}

void
newhandler (char *buf, struct conlist *t, struct conlist **cons)
{
  if (t->type != new)
    panic ("Wrong contype in newhandler", 1);
  if (strncmp (buf, "You there?", 10) == 0)
  {
    t->type = client;
    text_sock_write (t, "Yeah.\n");
    log_promo (t);
  } 
  else
    bozoize (t);
}

void
wannabehandler (char *buf, struct conlist *t, struct conlist **cons)
{
  if (t->type != wannabe)
    panic ("Wrong contype in wannabehandler", 1);
  if (strncmp (buf, "security is a myth", 18) == 0)
  {
    t->type = command;
    log_promo (t);
    text_sock_write (t, "Ready.\n");
  }
  else
  {
    t->type = client;
    log_promo (t);
    clienthandler (buf, t, cons);
  }
}

/* Connection types bozo and delete were created to avoid having subprograms */
/* that modify the connection list while the main program is scanning it or  */
/* having functions writing to conlist structs which have been freed.        */
/* The drawback is that the main program has to be persuaded to service them */
/* again even though no more messages have arrived.  So we add them to wrset.*/

/* This function has been enhanced to handle the case where a daemon is bozo'd
 * after it becomes fully operational.
 */
void
bozoize (struct conlist *t)
{
  int a = 0;
  if (t->type == daemon)
    a = 1;
  /* We mark it first to keep NETCHANGE from sending messages over the dead
  link. */
  t->type = bozo;
  FD_CLR (t->con, &fdset);
  FD_SET (t->con, &wrset);
  log_promo (t);
  if (a)
  {
    nc_del_column (t->name);
    fixup_qroutes ();
  }
}

/* This function is important.  It invokes the NETCHANGE routine for when
 * daemon links go down.
 */
void
delmark (struct conlist *t)
{
  int a = 0;
  if (t->type == daemon)
    a = 1;
  /* We mark it first to keep NETCHANGE from sending messages over the dead
  link. */
  t->type = delete;
  FD_CLR (t->con, &fdset);
  FD_SET (t->con, &wrset);
  log_promo (t);
  if (a)
  {
    nc_del_column (t->name);
    fixup_qroutes ();
    refigure_neighbor_costs ();
  }
}

static void
tease_bozo (struct conlist *t)
{
  char x[namlen+80];
  sprintf (x, "\n\n\nTo: Bozo@%s\n", t->name);
  text_sock_write (t, x);
  cat_text_file (t, "bozo.txt");
}

/* Default function to be called before a connection is deleted. */
void
no_notify (struct conlist *t)
{
  ;
}

void
service_cons (struct conlist **cons)
{
  int wasnuke, busy;
  struct conlist *t;
  wasnuke = nuke;

  /* This program performs internal buffering to break packets into separate
   * commands and to coalesce them into large data objects.  It is necessary to
   * loop here until all outstanding data have been read since data in our
   * internal buffers will not trigger the select in the main program.
   */

  /* Nukes are not guaranteed to work.  If flush_write does not succeed in
   * flushing the buffer the first time through, the nuke won't be sent.
   */

  do {
    busy = 0;

    /* This was moved here from the main loop because new connections were
     * being ignored while this do-loop stayed busy.
     */
    if (!nuke)
      accept_new_cons (cons);

    t = *cons;
    if (t == NULL && wasnuke)
    {
      printf ("Shutting down.\n");
      exit (0);
    }
    while (t != NULL)
    {
      if (t->type == delete)
      {
        struct conlist *u = t->next;
        /* Inform owner of connection that it is gone. */
        (*(t->notify)) (t);
        remove_con (cons, t);
        t = u;
      }
      else
      {
        if (t->type == bozo)
        {
          tease_bozo (t);
          delmark (t);
        }
        else if (wasnuke)
        {
          if (t->type == daemon)
          {
            conlog (t, "Sending NUKE to");
            text_sock_write (t, "nuke\n");
          }
          delmark (t);
        }
        else
        {
          if (t->sendlen)
            flush_write (t);
          /* flush_write can delmark t or promote connecting */
          if (t->type != delete && t->type != connecting)
          {
            /* Try to read */
            switch (get_message (&messbuf, &messbufsize, t)) {
            case -1:
              /* Connection has gone amok -- close it */
#ifndef logcons
              if (t->type == daemon)
#endif
              {
                conlog (t, "Dead connection from/to");
              }
              delmark (t);
              break;
            case 0:
              /* No traffic */
              break;
            case 1:
              /* Got a message */
              if ((!strlen(messbuf))&&(t->type == new))
              {
                /* If it's a blank line (just a CR), it shows up here as a
                 * null.  We want to bozo people that telnet in and just hit
                 * return, but other connections need to be able to send
                 * blank lines.
                 */
                bozoize (t);
              }
              else
              {
#ifdef logpackets
                printf ("Packet returned to main:  '%s'\n", messbuf);
                fflush (stdout);
#endif
                busy = 1;
                (*(t->handler)) (messbuf, t, cons);
              }
            }
          }
        }
        t = t->next;
      }
    }
    if (wasnuke)
      sleep (10);        /* Drain connections. */
  } while (busy || wasnuke);
}

/* Returns uniformly distributed random numbers in [0, 1). */
#ifdef norandom
#define RND_DENOM ((double)RAND_MAX+1.0)
#define random rand
#else
#define RND_DENOM 2147483648.0
#endif
float
rnd ()
{
  double temp = 2.0;
  while (temp < 0.0 || temp >= 1.0)   /* I've seen 1.0 before.... */
    temp = (double)(random ()) / RND_DENOM;
  return (float)temp;
}

main ()
{
  /* Try to act like a daemon. */
  if (fork ()) return 0;       /* Get away from parent. */
#ifndef nosetsid
  setsid ();                   /* Get away from controlling terminal. */
#endif
  /* nice (40); */             /* Generosity is its own punishment. */

  set_myhostname ();

#ifdef defdir
  if (chdir (defdir))
  {
    printf ("Cannot change directory to %s\n", defdir);
    puts ("Shutting down.");
    exit (-1);
  }
#endif
  {
    char t[namlen];
    /* getcwd seems to exist more often than getwd. */
    /* printf ("Default directory is %s.\n", getwd(t)); */
    printf ("Default directory is %s.\n", getcwd(t,namlen));
    fflush (stdout);
  }

  messbufsize = 1000;
  messbuf = safmalloc (messbufsize);
  FD_ZERO (&fdset);
  FD_ZERO (&wrset);
  nc_init ();                  /* Nullify routing tables */
  powerup ();                  /* Bring up the network */
  sig_setup ();                /* Signals */
  s_time = time((time_t *)0);
  printf ("Unetd activated on port %d at %s", inport_number,
    ctime(&s_time));
  fflush (stdout);

  /* Initialize resources */
  handle_query (NULL);

  while (1)
  {
    if (!nuke)
    {
      /* Random number generator seeded by net address controls when links
       * are attempted.  Repeated mutual bozoing due to duplicate links is
       * thus hopefully unlikely.
       */
      if ((rnd () < 1.0 - pow (1.0-1.0/(double)rel_sec,
      (double)(time((time_t *)0)-last_relink))) ||
      time((time_t *)0) - last_relink > (long)(0.75 * (float)maxidle))
        relink (&cons);
      last_relink = time((time_t *)0);

      /* Block until we have something to do.  We wake up every once in a */
      /* while just to make the rounds and give relink a chance to run. */
      {
        int ts;
        struct timeval tv;
        tfdset = fdset; twrset = wrset;
        tv.tv_sec = min ((long)(rnd()*(double)rel_sec), (long)maxidle);
        tv.tv_usec = 0;
        ts = select (FD_SETSIZE, &tfdset, &twrset, &tfdset, &tv);
        if (ts < 0)
        {
          perror ("select");
          fflush (stderr);
          /* DON'T PANIC!  Any old signal will cause this. */
        }
#ifdef markmesg
        if (ts == 0)
        {
          time_t ttime = time((time_t *)0);
          printf ("%.24s and all's well\n", ctime(&ttime));
          fflush (stdout);
        }
#endif
      }
      if (check_reset ())
        handle_query (NULL);
    }
    service_cons (&cons);
  }
}
