/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Header file for Unetd, contains configuration #defines.  Please be
 * aware that the debugging modes are not supported and may not even work.
 */

/* -------------------------------------------------------------------- */
/* You are encouraged to change the #defines in this section.           */
/* -------------------------------------------------------------------- */

/* If defdir is defined, chdir is called to set the default directory (or
 * current working directory) to the specified path.  The default directory
 * is expected to contain peers and bozo.txt.  FQDN is the only file which is
 * accessed prior to the changing of the directory.  #define defdir myhostname
 * is a nice way to deal with multiple servers sharing a filesystem.
 */
#define defdir myhostname

/* This is the number of bytes of memory, including overhead, that we will
 * use to cache data.  You can set it to zero if you want, but there is little
 * reason to do this.  On most relatively normal UNIX systems under moderate
 * load, the data cache will be swapped out to disk anyway.
 */
#define cachesize 1000000

/* setsid is a function which disassociates the caller from its
 * controlling terminal, which is a Good Thing.  Some compilers don't
 * have it.  Undefine this to use setsid.
 */
#undef nosetsid

/* Define this if you don't have random / srandom.  rand and srand, which tend
 * to be awful, will be used instead.
 */
#undef norandom

/* Define this if, for some strange reason, dirent.h is broken but sys/dir.h
 * is okay.
 */
#undef nodirent

/* The code for processing Boolean queries has two sections that can be
 * very CPU-intensive.  Leaving this undefined disables one of those sections
 * with few ill effects.  Defining this may have many ill effects if your CPU
 * is not blazingly fast.
 */
#undef slow_Booleans

/* Connect doesn't go right with O_NONBLOCK on BSD Suns.  Nothing goes right
 * with O_NDELAY on the SYSV HP we have to use.  All other machines work
 * fine with either one.  If read returns 0 instead of 1 on your machine
 * when O_NDELAY is set and there is no data on a socket, #define __hpux
 * here.
 */

/* -------------------------------------------------------------------- */
/* The less configurable #defines come after the #includes.             */
/* -------------------------------------------------------------------- */

/* #include <everything.h> and get it over with */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef nodirent
#include <sys/dir.h>
#else
#include <dirent.h>
#endif
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

#ifdef sig_debug
#include "leak.h"
#endif

/* Delete these if they are redefs and cause problems */
#define max(a,b) (a < b ? b : a)
#define min(a,b) (a > b ? b : a)

/* Number of signals, should be set by signal.h but not always is */
#ifndef NSIG
#define NSIG 32
#endif

/* Some compilers actually don't define O_NONBLOCK.  O_NDELAY is the same
 * on some machines; on others, it wreaks havoc with netutil.c.  Pray
 * that this has the desired effect!
 */
#ifndef O_NONBLOCK
#define O_NONBLOCK O_NDELAY
#endif

/* -------------------------------------------------------------------- */
/* Tune at your own risk                                                */
/* -------------------------------------------------------------------- */

/* The desired cache turnover.  Higher numbers mean less turnover.  A given
 * value means that we want to have things being replaced on the average
 * after that many queries have been processed since the replica was created.
 */
#define best_turnover 40

/* More cache stuff. */
#define warm_hint 10
#define cold_hint 0
#define fudge_factor 0.001

/* Query routing stuff.  These settings max out at 100. */
#define agefac 0.99
#define boostfac 1.0

/* This is important.  If the lag for reaching a host exceeds
 * this amount, we consider the host to be unreachable.  This
 * must be set to a reasonable number so that NETCHANGE loops
 * will be terminated eventually.  This value MUST be less than
 * 256,000,000 (256 seconds) because the timer rolls over at
 * that value.
 *
 * Note that it is possible for a bad link to come back with a lag
 * which exceeds maxlag without the link going down.  A host is only
 * considered unreachable if there does not exist a route to the host
 * such that the total lag is less than maxlag.
 */
#define maxlag 20000000

/* This is also important.  Some machines don't really support microsecond
 * accuracy, so they fake it by incrementing the time by 1 each time you
 * call and then resetting it at each clock tick.  This can cause a ping time
 * of 1 microsecond if the actual time is within one clock tick.  If the ping
 * time is less than minlag, we use minlag.  10000 = 1/100 second.
 */
#define minlag 5000

/* After this many seconds with no activity, a connection is assumed dead. */
#define maxidle 7200

/* Define if you want messages in the log akin to MARK in syslog */
#undef markmesg

/* Define this to record netchange messages individually. */
#undef lognetc

/* Define this to log all sorts of information about connections. */
#undef logcons

/* Number of seconds on the average between relinks.  Note that maxidle is
 * also taken into account when deciding when to relink.
 */
#define rel_sec 1800

/* These control the hashing of hostnames in the distance table.  If hashstat
 * shows a lot of buckets with multiple occupancy, you can increase numbuckets.
 * If mult is not extremely prime with respect to numbuckets, the distribution
 * of hostnames into buckets will be uneven.
 */
#define numbuckets 41
#define mult 107

/* Query messages which take this many hops are returned. */
#define TTL_query 25
/* Response messages which take more than this many hops are dropped. */
#define TTL_resp 25

/* -------------------------------------------------------------------- */
/* Don't bother reading the rest unless you have problems               */
/* -------------------------------------------------------------------- */

/* Maximum length of hostnames, filenames, short strings in general */
#define namlen 256

/* Debug option to log all data as they are received.  Warning:  binary data
 * are not filtered out.
 */
#undef logpackets

/* Log queries received from users */
#define logqueries

/* Ultra-detailed explanations of query processing */
#undef debugquery

/* Explain why each datum is cached or not cached. */
#define debugcache

/* No nukes is good news. */
#undef enable_nukes

/* -------------------------------------------------------------------- */
/* Show's over                                                          */
/* -------------------------------------------------------------------- */

/* Connection types:
 *     connecting       Waiting for outbound connection to connect
 *     new              New inbound connection
 *     bozo             Connection being bozo'd
 *     near_daemon      In process of setting up daemon connection
 *     daemon           Daemon connection ("link")
 *     daemon_oob       Not yet implemented separate connection for large data
 *     client           Client connection
 *     command          So-called command channel
 *     wannabe          Client in process of setting up command channel
 *     delete           Connection being deleted
 *     resource         Resource (actually a pipe!)
 *     generic          Deprecated.
 */
typedef enum {resource, connecting, new, bozo, near_daemon, daemon, daemon_oob,
client, command, wannabe, delete, generic} contype;

typedef enum {question, answer, garbage} dmestype;

/* A data object is stored as a linked list of text messages, some of which
 * have a block of binary data associated with them.  This is also how they are
 * transmitted.
 */
struct dobj {
  char *text;
  char *bindata;
  long binlen;
  struct dobj *next;
};

struct conlist {
  struct conlist *next;        /* Linked list */

/* Connection information */

  int con;                     /* fd for input (sockets and pipes) */
  int out_con;                 /* fd for output (same as con for sockets) */
  void (*handler) ();          /* Where to process received messages */
  void (*notify) ();           /* Notification of deletion */
  unsigned long snum;          /* Serial number of connection */
  contype type;                /* Type of connection. */
  contype n_type;              /* Type to promote to after connect */
  struct sockaddr addr;        /* Info on peer */
  char *name;                  /* Name of peer. */
  char *number;                /* Numerical address of peer. */
  int pingflag;                /* Are we the ping-er or the ping-ee */
  long lag;                    /* Round trip ping time */
  long lastrecv;               /* Timestamp of last receive */

/* Input-related stuff */

  char *inbuffer;              /* Generic input buffer */
  long inbuflen;               /* How long is it */
  long inmemsize;              /* Size of memory block >= inbuflen */

  char *sidetrack;             /* Temporary holding tank for header line */
  char *bindata;               /* Temporary holding tank for binary data */
  long binsize;                /* Size of bindata, if any */
  struct dobj *datum;          /* Head of data object */
  struct dobj *enddatum;       /* Tail of data object */
  int blanks;                  /* 3 blank lines = bozo'd. */

/* Output-related stuff */

  char *yettosend;             /* Generic output buffer */
  long sendlen;                /* How long is it */
  long sendmemsize;            /* Size of memory block >= sendlen */

  char forsite[namlen];        /* Name of source/destination site */
  unsigned long forsnum;       /* Serial number of source connection */
  dmestype dmes;               /* Is this a question or an answer? */
};

struct route_entry {
  char dest[namlen];
  long distance;
  char nexthop[namlen];
  struct route_entry *next;
};

struct qrrow {
  struct conlist *nexthop;
  float factor;
  struct qrrow *next;
};

struct qrtable {
  char *data_type;
  struct qrrow *row;
  struct qrtable *next;
};

/* The "real" declarations for these are in Unetd.c */
extern int errno;
extern int insocket, nuke, inport_number, nfds;
extern long netccount, pingrqcount, mrt, noncracc, noncrdel,
  pingrscount, lagcount, querycount, responsecount, anscount, qacc, qdel,
  cachedel;
extern time_t s_time, last_relink;
extern fd_set fdset, wrset, tfdset, twrset;
extern unsigned long serialnum, c_hops, c_links, c_cache;
extern char myhostname[];
extern struct conlist *cons;
extern struct qrtable *qroutes;
extern long freespace, myaddr, neighborlag, unrealtime;

/* portnums.c */
extern void reset_port_series ();
extern int next_port_number ();
extern int photogenesis ();

/* netutil.c */
extern void refigure_neighbor_costs ();
extern void flush_write ();
extern void ping ();
extern void powerup ();
extern void text_sock_write ();
extern void bin_sock_write ();
extern void accept_new_cons ();
extern void remove_con ();
extern int get_message ();
extern int mklink ();
extern void cat_text_file ();
extern int duplink ();
extern void relink ();
extern struct conlist *findcon ();
extern struct conlist *gen_connect ();

/* Unetd.c */
extern char *title ();
extern void rehandle ();
extern void bitbucket ();
extern void connectinghandler ();
extern void newhandler ();
extern void wannabehandler ();
extern void bozoize ();
extern void delmark ();
extern void no_notify ();
extern float rnd ();

/* lclient.c */
extern void clienthandler ();
extern void addline ();

/* lcommand.c */
extern void comhandler ();

/* ldaemon.c */
extern void daemonhandler ();
extern void handle_response ();
extern void initiate_response ();

/* netchang.c */
struct route_entry *routeto ();
extern void nc_new_lag ();
extern void nc_send_table ();
extern void nc_init ();
extern void nc_hashstat ();
extern void nc_add_column ();
extern void nc_del_column ();
extern void nc_netchange ();

/* lnear.c */
extern void nearhandler ();

/* control.c */
extern void conlog ();
extern void timlog ();
extern void log_promo ();
extern unsigned long microtime ();
extern void panic ();

/* signals.c */
extern void sig_setup ();
extern int broke_pipe ();
extern void unbreak_pipe ();
extern int check_reset ();

/* strutil.c and strfuns.c */
extern char *cstrstr ();
extern char *sstrstr ();
extern void strdowncase ();
extern int cstrcmp ();
extern int cstrncmp ();
extern int sstrncmp ();
extern char *noeol ();
extern char *nopunct ();
extern void append_to_queue ();
extern void shorten_queue ();
extern char *strclone ();
extern void free_dobj ();
extern char *startoftext ();
extern char *tstartoftext ();
extern int sremove ();
extern void sremall ();
extern char *strword ();
extern char *dejunk ();
extern char *translate ();
extern struct dobj *filter_clone_dobj ();
extern struct dobj *clone_dobj ();
extern int wrdmatches ();
extern int struncump ();
extern void *safmalloc ();

/* query.c */
extern void handle_query ();
extern void qroute_hook ();
extern void fixup_qroutes ();
extern void rebound ();
extern void reply ();
extern int forward ();
extern void forward_query ();
extern void fail_query ();
extern void list_qroutes ();

/* classify.c */
extern char *classify ();
extern char *describe ();
extern void add_class ();
extern void add_descript ();
extern void change_class ();

/* cache.c */
extern void dump_cache ();
extern void cache_hook ();
extern void *res_cache ();

/* oids.c */
extern char *find_oid ();
extern void identify_dobj ();
extern int version_blind_id_cmp ();

/* resource.c */
extern void resourcehandler ();
extern void connect_resources ();
