/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This is a minimal alibi client.
 *
 *    Usage:  alibi server.host.fqdn.whatever
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define namlen 80

int server;
char servname [namlen];

/* Get rid of trailing CR's and LF's */
char *
noeol (char *s)
{
  char *a;
  a = &(s[strlen(s)]);
  while (a > s)
    if (*(a-1) == '\n' || *(a-1) == '\r')
      *(--a) = '\0';
    else
      break;
  return s;
}

char *
tstartoftext (char *u)
{
  char *a;
  if (!u)
    return NULL;
  a = strchr (u, ' ');
  if (!a)
    return (u + strlen (u));
  while (*a == ' ')
    a++;
  return a;
}

/* Get one character from the server. */
char
sgetc ()
{
  char t;
  int l=0;
  while (!l)
  {
    l = read (server, &t, 1);
    if (l <= -1)
    {
      puts ("Error on read from server -- shutting down.");
      exit (-1);
    }
  }
  return t;
}

/* Send one character to the server. */
void
sputc (char t)
{
  int l=0;
  while (!l)
  {
    l = write (server, &t, 1);
    if (l <= -1)
    {
      puts ("Error on write to server -- shutting down.");
      exit (-1);
    }
  }
}

/* Get a line of text */
char *
get_text ()
{
  static char buf[10000];
  char c='a';
  int l=0;
  while (c != '\n')
  {
    if (l >= 10000)
    {
      puts ("Input buffer overflow.  Shutting down.");
      exit (-1);
    }
    c = sgetc ();
    buf[l++] = c;
  }
  buf[l] = '\0';
  return buf;
}

/* Write a line of text (adding \r\n) */
void
send_message (char *mess)
{
  int a;
  for (a=0;a<strlen(mess);a++)
    sputc (mess[a]);
  sputc ('\r');
  sputc ('\n');
}

/* Connect to the server. */
void
startup ()
{
  struct sockaddr_in sou;
  struct hostent *t;

  t = (struct hostent *) gethostbyname (servname);
  if (t == NULL)
  {
    puts ("gethostbyname failed.  In other words:  unknown host!");
    exit (-1);
  }
  memset ((char *) &sou, '\0', sizeof sou);
  memcpy ((char *) &sou.sin_addr, t->h_addr, t->h_length);
  sou.sin_family = (short)(t->h_addrtype);
  sou.sin_port = (u_short) (htons ((u_short)13592));

  server = socket (t->h_addrtype, SOCK_STREAM, 0);
  if (server == -1)
  {
    puts ("Could not open socket.  This machine may be getting hosed.");
    exit (-1);
  }

  if (connect (server, (struct sockaddr *)(&sou), sizeof sou))
  {
    puts ("Could not connect.  Try a different server.");
    exit (-1);
  }

  send_message ("You there?");
  if (strcmp (get_text (), "Yeah.\r\n"))
  {
    puts ("Server acting weird.  Shutting down.");
    exit (-1);
  }
}

/* The code: (bits)
 *   1 = last message in response
 *   2 = start of data block
 */
void
cope_with_reply (char *oid)
{
  char *a, *u;
  int b=0;
  oid[0] = '\0';
  while (!(b&1))
  {
    a = get_text ();
    printf ("%s", (u=tstartoftext (a)));
    if (sscanf (a, "%d", &b) != 1)
    {
      puts ("Server acting weird.  Shutting down.");
      exit (-1);
/*
      b=0;                 Use this instead if you just want to print out
      continue;                the weird stuff and hope for the best.
*/
    }
    if (!(strncmp (u, "ID: ", 4)))
    {
      strcpy (oid, "[");
      strcat (oid, u+4);
      noeol (oid);
      strcat (oid, "]");
    }
    if (b&2)
    {
      unsigned long length, junk1;
      char filename[namlen], junk3[namlen];
      FILE *fp;
      if (sscanf (a, "%d Binary data follows ... %lu bytes, filename %s",
        &junk1, &length, filename) != 3)
      {
        puts ("Error parsing Binary data line.");
        exit (-1);
      }
      noeol (filename);
      printf ("\nWarning:  existing files will be appended without further warning\n");
      printf ("Choose filename in which to save response [%s]: ", filename);
      gets (junk3);
      if (strlen (junk3))
        strcpy (filename, junk3);
      if (!(fp = fopen (filename, "ab")))
      {
        puts ("Could not open file -- trashing response");
        if (!(fp = fopen ("/dev/null", "ab")))
        {
          puts ("Could not open /dev/null -- shutting down.");
          exit (-1);
        }
      }
      puts ("Saving response....");
/*
      for (junk1=0;junk1<length;junk1++)
        if (fputc (sgetc (), fp) == EOF)
        {
          puts ("Error on write to file -- shutting down.");
          exit (-1);
        }
*/
      /* This is faster. */
      junk1=0;
      while (junk1<length)
      {
        int temp;
        char buf[4096];
        if (length-junk1 < 4096)
          temp = read (server, buf, length-junk1);
        else
          temp = read (server, buf, 4096);
        junk1 += temp;
        /* They say never to test for an error you don't know how to handle,
         * so we'll assume that this write succeeds.
         */
        fwrite (buf, 1, temp, fp);
      }
      if (fclose (fp) == EOF)
      {
        puts ("Error on fclose?!?  Sheesh!  Shutting down.");
        exit (-1);
      }
    }
  }
}

void
print_help ()
{
  printf ("\n");
  puts ("HELP for Minimal Alibi Client");
  puts ("-----------------------------");
  puts ("This client provides few user-friendly services.  You must enter your queries");
  puts ("in the correct postorder format or die horribly.  Some examples follow.");
  puts ("  (yak gnu)          Match yak and gnu (database level logic).");
  puts ("  (yak)(gnu)&        Nearly the same (intersection at server level instead).");
  puts ("  (yak)(gnu)|        Match either yak or gnu.");
  puts ("  (yak)(gnu)~&       Match yak as long as no gnus involved.");
  puts ("  (yak)(gnu)|(bison)&    Match bison and either yak or gnu.");
  puts ("You can also say 'more' to automatically construct a query that will try to");
  puts ("get something else like what you just got.");
  puts ("\nIf you retrieve something that you don't want, save it to /dev/null.");
  puts ("\nDon't expect a quick response -- file transfers take a while!");
  puts ("\n(C) 1994 David Flater.");
}

void
doit (char *buff, char *prevquery, char *oid)
{
  send_message (buff);
  puts ("\nWaiting for reply....");
  cope_with_reply (oid);
  if (oid[0] == '[')
  {
    strcat (prevquery, oid);
    strcat (prevquery,"~&");
  }
  else
    prevquery[0] = '\0';
}

void
mainloop ()
{
  char buff[10000];
  char prevquery[10000];
  char oid[256];
  prevquery[0] = '\0';
  while (1)
  {
    puts ("\nEnter a query now, or enter 'quit' to quit.");
    gets (buff);
    if (!(strcmp (buff, "quit")))
      return;
    if (!(strcmp (buff, "more")))
    {
      if (prevquery[0] == '\0')
      {
        puts ("No more.  Start over.");
        continue;
      }
      strcpy (buff, prevquery);
      puts (prevquery);
      doit (buff, prevquery, oid);
      continue;
    }
    if (buff[0] != '(')
    {
      print_help ();
      continue;
    }
    if (strlen (buff))
    {
      strcpy (prevquery, buff);
      doit (buff, prevquery, oid);
      continue;
    }
  }
}

main (int argc, char **argv)
{
  puts ("Minimal alibi client v1.0, Copyright (C) 1994 David W. Flater.");
  puts ("This client comes with ABSOLUTELY NO WARRANTY.  This is free");
  puts ("software, and you are welcome to redistribute it under certain");
  puts ("conditions.  Please refer to the Gnu General Public License");
  puts ("provided in the Alibi source code distribution for details.\n");
  if (argc !=2)
  {
    puts ("Usage:  alibi server_hostname (just like telnet)");
    puts ("Use localhost if you can; otherwise pick a nearby server.");
    exit (-1);
  }
  strcpy (servname, argv[1]);
  startup ();
  mainloop ();
  close (server);
  exit (0);
}
