/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* control.c contains functions for logging, timing, and bombing out. */

#include "alibi.h"

/* The second parameter was included to support recovery for nonfatal errors.
 * Such recovery was never implemented; most nonfatal errors are already
 * handled some other way.
 */
void
panic (char *reason, int fatal)
{
  fflush (stdout);
  fflush (stderr);

  fprintf (stderr, "Panic.\n");
  if (reason)
    fprintf (stderr, "Reason:  %s\n", reason);
  else
    fprintf (stderr, "No reason given.\n");
  fprintf (stderr, "Shutting down.\n");
  fflush (stderr);
  exit (1);
}

/* Provide timestamps for pings and that sort of thing.  If your machine
 *  doesn't have gettimeofday, fake it somehow.
 */
unsigned long
microtime ()
{
  struct timeval tp;
  if (gettimeofday (&tp, (struct timezone *)0))
    panic ("gettimeofday failed!", 1);
  return (tp.tv_sec & 0xff) * 1000000 + tp.tv_usec;
}

void
timlog (char *descript)
{
  time_t contime = time((time_t *)0);
  printf ("%s %.24s\n", descript, ctime (&contime));
  fflush (stdout);
}

void
conlog (struct conlist *u, char *descript)
{
  time_t contime = time((time_t *)0);
  struct sockaddr_in *t = (struct sockaddr_in *)(&(u->addr));
  if (u->number)
    printf ("%s %s.%hu (%s) %.24s\n", descript, u->number, ntohs (t->sin_port),
      u->name, ctime (&contime));
  else
    printf ("%s %s %.24s\n", descript, u->name, ctime (&contime));
  fflush (stdout);
}

void
log_promo (struct conlist *t)
{
  char x[80];
#ifndef logcons
  if (t->type == bozo)
#endif
  {
    sprintf (x, "Type set to %s:", title (t->type));
    conlog (t, x);
  }
  rehandle (t);    /* Very important */
}
