/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  redgar.c
 *
 *  Derived from rblobs, this mediator accesses the SEC's EDGAR database.
 *  The script ftp.sh must be present for this to work.
 *
 *  When they mess with the format of EDGAR, this will have to be updated.
 */

#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>

void
reset_blobs ();
void
fetch ();
void
print ();
void
do_query ();
void
ditch ();
void
net_fail ();

/* EDGAR-specific stuff */
#define ftp_script   "/home/flater/alibi/example_resources/ftp.sh"
#define ftp_recv     "/tmp/ftp_recv"
#define unc_recv     "/tmp/ftp_uncompressed"
#define index_file   "edgar/full-index/company.Z"

/* Approx. number of OID's returned for a () query */
#define universe_size 30

/* This directory only matters for the named pipes. */
#define defdir "/home/flater/albin/zing.ncsl.nist.gov"

int infd, outfd;
long myaddr=0, idxsize=0;
double randmax;  /* RAND_MAX + 1 */

/* The following is the class of data which is prepended to the terms in the
 * directory path to form the classification of each blob.
 */
#define blob_class "government USA SEC EDGAR"

struct bdex {
  char *file;
  char *descript;
  struct bdex *next;
};

struct bdex *vindex = NULL;

/*
 *  FTP a file.
 */
void
get_file (char *fname)
{
  char buf[1000];
  unlink (ftp_recv);
  sprintf (buf, "%s %s > /dev/null 2>&1", ftp_script, fname);
  system (buf);
}

void
shutdown ()
{
  print ("BYE\r\n");
  /* Let Unetd read the BYE */
  sleep (10);
  exit (-1);
}

void
timlog (char *a)
{
  print ("CONLOG ");
  print (a);
  print ("\r\n");
}

void
panic (char *why, int a)
{
  timlog (why);
  shutdown ();
}

#include "strfuns.c"

void
create_pipe (int *fd, char *name)
{
  *fd = -1;
  if (mknod (name, S_IFIFO | S_IREAD | S_IWRITE, NULL))
  {
    perror ("mknod");
    exit (-1);
  }
  *fd = open (name, O_RDWR, NULL);
  if (*fd == -1)
  {
    perror ("open");
    exit (-1);
  }
}

void
print (char *m)
{
  write (outfd, m, strlen (m));
}

/* Get one character from the server. */
char
sgetc ()
{
  char t;
  int l=0;
  while (!l)
  {
    l = read (infd, &t, 1);
    if (l <= -1 && errno == EINTR)
    {
      l = 0;
      continue;
    }
    if (l <= -1)
      panic ("Error on read from server -- shutting down.", 1);
  }
  return t;
}

/* Get a line of text */
char *
get_text ()
{
  static char buf[10000];
  char c='a';
  int l=0;
  while (c != '\n')
  {
    if (l >= 10000)
      panic ("Input buffer overflow.  Shutting down.", 1);
    c = sgetc ();
    buf[l++] = c;
  }
  buf[l] = '\0';
  return buf;
}

void
handler (char *m)
{
  noeol (m);
  if (strncmp (m, "RESET", 5) == 0)
  {
    if (sscanf (m, "RESET %ld", &myaddr) != 1)
    {
      timlog (m);
      timlog ("RESET didn't provide address");
    }
    reset_blobs ();
  }
  else if (m[0] == '(')
  {
    if (!myaddr)    /* Not ready yet */
    {
      timlog ("Can't answer queries until RESET provides address");
      print ("MATCHES\r\n");
    }
    else
      do_query (m);
  }
  else if (strncmp (m, "FETCH [", 7) == 0)
    fetch (m+6);
  else
  {
    print ("CONLOG What is all this garbage? (see below)\r\nCONLOG ");
    print (m);
    print ("\r\n");
  }
}

/* Inform Unetd that we are leaving. */
void
sig_shutdown (int sig)
{
  printf ("Received unexpected signal %d -- shutting down.\n", sig);
  shutdown ();
}

/* Administrator can manually send a SIGPIPE if Unetd crashes to keep from
 * having to restart the resource.
 */
void
newpipes ()
{
  signal (SIGPIPE, newpipes);
  close (infd);
  close (outfd);
  create_pipe (&infd, "wres_EDGAR");
  create_pipe (&outfd, "rres_EDGAR");
  print ("CONLOG new pipes\r\n");
}

void
do_sigs ()
{
  int a;
  for (a=1;a<NSIG;a++)
    if (a != SIGCONT)
      signal (a, sig_shutdown);
  signal (SIGPIPE, newpipes);
}

/******** End generic resource code, begin blob-specific stuff *******/

/* I got sick of compilers that didn't define RAND_MAX. */
void
rand_setup ()
{
  long temp,max=0,a;
  srand (time (NULL));
  for (a=0;a<100;a++)
  {
    temp = rand ();
    if (temp > max)
      max = temp;
  }
  if (max < 32768)
    randmax = 32768.0;
  else
    randmax = 2147483648.0;
}

void
wipe_index ()
{
  while (vindex)
  {
    struct bdex *t;
    if (vindex->descript)
      free (vindex->descript);
    if (vindex->file)
      free (vindex->file);
    t = vindex->next;
    free (vindex);
    vindex = t;
  }
  idxsize = 0;
}

struct bdex *
srch_by_fname (char *fname)
{
  struct bdex *t = vindex;
  while (t)
  {
    if (!(strcmp (fname, t->file)))
      return t;
    t = t->next;
  }
  return NULL;
}

/* This function has been made EDGAR-specific and may need alterations when
 * they change things around.
 */
void
get_index ()
{
  char buf[1000];
  FILE *fp;
  get_file (index_file);
  sprintf (buf, "uncompress < %s > %s", ftp_recv, unc_recv);
  system (buf);
  unlink (ftp_recv);
  if (!(fp = fopen (unc_recv, "rt")))
  {
    timlog ("Couldn't open index file");
    net_fail ();
    return;
  }

  /* Skip junk */
  do {
    if (!(fgets (buf, 1000, fp)))
      panic ("Failure finding start of index file", 1);
  } while (strncmp (buf, "----------", 10));

  while (fgets (buf, 1000, fp))
  {
    struct bdex *t;
    noeol (buf);
    {  /* Ditch trailing blanks in an unsafe manner */
      char *b;
      b = buf + strlen (buf) -1;
      while (*b == ' ')
        *(b--) = '\0';
    }
    t = (struct bdex *) safmalloc (sizeof (struct bdex));

    /* WARNING:  COLUMN-ORIENTED FORMAT!!!  */

    t->file = strclone (buf+98);
    t->descript = strclone (buf);
    t->next = vindex;
    vindex = t;
    nopunct (t->descript);
    idxsize++;
  }
  fclose (fp);
  unlink (unc_recv);
}

void
reset_blobs ()
{
  print ("CONLOG Blobs resource reset\r\n");
  wipe_index ();
  get_index ();
  print ("READY\r\n");
#ifdef blob_class
  print ("MYTYPE ");
  print (blob_class);
  print ("\r\n");
#endif
}

void
send_fetch_oid (struct bdex *c)
{
  char temp[namlen*2];
  sprintf (temp, "[%s %ld %lu]", c->file, myaddr, (unsigned long)0);
  print (temp);
}

void
do_query (char *query)
{
  struct bdex *c;
  nopunct (query);
  if (strspn (query, " ") == strlen (query))
    query[0] = '\0';
  print ("MATCHES ");

  c = vindex;
  while (c)
  {
    int tmatches, tlength;
    if (strlen (query))
    {
      if (c->descript)
      {
        tmatches = wrdmatches (c->descript, query, &tlength);
        if (tmatches == tlength)
          send_fetch_oid (c);
      }
    }
    else
    {
      /* Universal query -- limit size of response */
      if ((double)rand () / randmax < (double)universe_size/(double)idxsize)
        send_fetch_oid (c);
    }
    c = c->next;
  }
  print ("\r\n");
}

/* Remove an index entry for a non-existent or problematic file. */
#if 0
void
ditch (char *f2ditch)
{
  struct bdex *p, *q;
  if (!(strcmp (vindex->file, f2ditch)))
  {
    p = vindex;
    vindex = p->next;
    free (p->file);
    if (p->descript)
      free (p->descript);
    free (p);
    idxsize--;
    return;
  }
  p = vindex;
  q = p->next;
  while (q)
  {
    if (!(strcmp (q->file, f2ditch)))
    {
      p->next = q->next;
      free (q->file);
      if (q->descript)
        free (q->descript);
      free (q);
      idxsize--;
      return;
    }
    p = q;
    q = q->next;
  }
}
#endif

/* Respond to a FETCH from Unetd.
 */
void
fetch (char *oid)
{
  struct bdex *hit;
  struct stat s;
  char fname[namlen], newoid[namlen*2];
  long looper;
  FILE *fp;
  sscanf (oid, "[%s", fname);
  hit = srch_by_fname (fname);
  if (!hit)
  {
    sprintf (newoid, "Can't find '%s'", fname);
    timlog (newoid);
    print ("OOPS\r\n");
    return;
  }
  get_file (fname);
  if (stat (ftp_recv, &s) == -1)
  {
    sprintf (newoid, "Can't stat '%s'", fname);
    timlog (newoid);
    print ("OOPS\r\n");
    /* ditch (fname); */     /* Assume it's a transitory failure */
    return;
  }
  fp = fopen (ftp_recv, "rb");
  if (!fp)
  {
    sprintf (newoid, "Can't open '%s'", fname);
    timlog (newoid);
    print ("OOPS\r\n");
    /* ditch (fname); */     /* Assume it's a transitory failure */
    return;
  }
  print ("0 The following datum was provided by the SEC's EDGAR database via FTP.");
  print ("\r\n0 ID: ");
  sprintf (newoid, "%s %ld %lu", fname, myaddr, (unsigned long)0);
  print (newoid);
  print ("\r\n0 Class: ");
  print (blob_class);
  print ("\r\n0 Description: ");
  print (hit->descript);
  print ("\r\n");
  sprintf (newoid, "3 Binary data follows ... %ld bytes, filename %s\r\n",
    s.st_size, strrchr(fname,'/') ? strrchr(fname,'/')+1 : fname);
  print (newoid);
/*
  for (looper=0;looper<s.st_size;looper++)
  {
    unsigned char temp;
    temp = (unsigned char) fgetc (fp);
    write (outfd, &temp, 1);
  }
*/
  /* This is faster. */
  looper=0;
  while (looper<s.st_size)
  {
    int temp;
    char buf[4096];
    temp = fread (buf, 1, 4096, fp);
    looper += temp;
    /* They say never to test for an error you don't know how to handle,
     * so we'll assume that this write succeeds.
     */
    write (outfd, buf, temp);
  }
  fclose (fp);
  unlink (ftp_recv);
}

/* Go away and try again later, for example, if couldn't read the index. */
void
net_fail ()
{
  print ("BYE\r\n");
  /* Let Unetd read the BYE */
  sleep (10);
  newpipes ();
  myaddr = 0;
}

main ()
{
#ifdef defdir
  if (chdir (defdir))
  {
    printf ("Cannot change directory to %s\n", defdir);
    puts ("Shutting down.");
    exit (-1);
  }
#endif
  if (fork ()) return 0;       /* Get away from parent. */
#ifndef nosetsid
  setsid ();                   /* Get away from controlling terminal. */
#endif
  do_sigs ();
  rand_setup ();
  create_pipe (&infd, "wres_EDGAR");
  create_pipe (&outfd, "rres_EDGAR");
  while (1)
    handler (get_text ());
}
