/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  rfips55-2
 *
 *  Derived from redgar (which was derived from rblobs), this mediator
 *  answers queries on a geographical database conforming to FIPS 55-2.
 *  The blobs-based variable names have long since ceased to have meaning.
 */

#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>

void
reset_blobs ();
void
fetch ();
void
print ();
void
do_query ();

#define database "/home/flater/henry_tom_gis/va.txt.gz"
#define unc_tmp  "/tmp/va.txt"

/* Approx. number of OID's returned for a () query */
#define universe_size 30

/* This directory only matters for the named pipes. */
#define defdir "/home/flater/albin/sunset.ncsl.nist.gov"

int infd, outfd;
long myaddr=0, idxsize=0;
double randmax;  /* RAND_MAX + 1 */

#define blob_class "government USA census geography VA"

struct bdex {
  char *file;            /* FIPS record */
  char *descript;        /* Unpunctuated place name */
  int ssn;               /* Serial number */
  struct bdex *next;
};

struct bdex *vindex = NULL;

void
shutdown ()
{
  print ("BYE\r\n");
  /* Let Unetd read the BYE */
  sleep (10);
  exit (-1);
}

void
timlog (char *a)
{
  print ("CONLOG ");
  print (a);
  print ("\r\n");
}

void
panic (char *why, int a)
{
  timlog (why);
  shutdown ();
}

#include "strfuns.c"

void
create_pipe (int *fd, char *name)
{
  *fd = -1;
  if (mknod (name, S_IFIFO | S_IREAD | S_IWRITE, NULL))
  {
    perror ("mknod");
    exit (-1);
  }
  *fd = open (name, O_RDWR, NULL);
  if (*fd == -1)
  {
    perror ("open");
    exit (-1);
  }
}

void
print (char *m)
{
  write (outfd, m, strlen (m));
}

/* Get one character from the server. */
char
sgetc ()
{
  char t;
  int l=0;
  while (!l)
  {
    l = read (infd, &t, 1);
    if (l <= -1 && errno == EINTR)
    {
      l = 0;
      continue;
    }
    if (l <= -1)
      panic ("Error on read from server -- shutting down.", 1);
  }
  return t;
}

/* Get a line of text */
char *
get_text ()
{
  static char buf[10000];
  char c='a';
  int l=0;
  while (c != '\n')
  {
    if (l >= 10000)
      panic ("Input buffer overflow.  Shutting down.", 1);
    c = sgetc ();
    buf[l++] = c;
  }
  buf[l] = '\0';
  return buf;
}

void
handler (char *m)
{
  noeol (m);
  if (strncmp (m, "RESET", 5) == 0)
  {
    if (sscanf (m, "RESET %ld", &myaddr) != 1)
    {
      timlog (m);
      timlog ("RESET didn't provide address");
    }
    reset_blobs ();
  }
  else if (m[0] == '(')
  {
    if (!myaddr)    /* Not ready yet */
    {
      timlog ("Can't answer queries until RESET provides address");
      print ("MATCHES\r\n");
    }
    else
      do_query (m);
  }
  else if (strncmp (m, "FETCH [", 7) == 0)
    fetch (m+6);
  else
  {
    print ("CONLOG What is all this garbage? (see below)\r\nCONLOG ");
    print (m);
    print ("\r\n");
  }
}

/* Inform Unetd that we are leaving. */
void
sig_shutdown (int sig)
{
  printf ("Received unexpected signal %d -- shutting down.\n", sig);
  shutdown ();
}

/* Administrator can manually send a SIGPIPE if Unetd crashes to keep from
 * having to restart the resource.
 */
void
newpipes ()
{
  signal (SIGPIPE, newpipes);
  close (infd);
  close (outfd);
  create_pipe (&infd, "wres_FIPS_55-2");
  create_pipe (&outfd, "rres_FIPS_55-2");
  print ("CONLOG new pipes\r\n");
}

void
do_sigs ()
{
  int a;
  for (a=1;a<NSIG;a++)
    if (a != SIGCONT && a != SIGCHLD)
      signal (a, sig_shutdown);
  signal (SIGPIPE, newpipes);
}

/******** End generic resource code, begin blob-specific stuff *******/

/* I got sick of compilers that didn't define RAND_MAX. */
void
rand_setup ()
{
  long temp,max=0,a;
  srand (time (NULL));
  for (a=0;a<100;a++)
  {
    temp = rand ();
    if (temp > max)
      max = temp;
  }
  if (max < 32768)
    randmax = 32768.0;
  else
    randmax = 2147483648.0;
}

void
wipe_index ()
{
  while (vindex)
  {
    struct bdex *t;
    if (vindex->descript)
      free (vindex->descript);
    if (vindex->file)
      free (vindex->file);
    t = vindex->next;
    free (vindex);
    vindex = t;
  }
  idxsize = 0;
}

struct bdex *
srch_by_ssn (int n)
{
  struct bdex *t = vindex;
  while (t)
  {
    if (n == t->ssn)
      return t;
    t = t->next;
  }
  return NULL;
}

/* This function has been made FIPS-specific.
 */
void
get_index ()
{
  char buf[140];
  FILE *fp;
  sprintf (buf, "gzip -dc < %s > %s", database, unc_tmp);
  system (buf);
  if (!(fp = fopen (unc_tmp, "rt")))
    panic ("Couldn't open database file", 1);
  unlink (unc_tmp);
  while (fgets (buf, 140, fp))
  {
    struct bdex *t;
    noeol (buf);
    t = (struct bdex *) safmalloc (sizeof (struct bdex));
    t->file = strclone (buf);
    buf[67] = '\0';
    t->descript = strclone (buf+15);
    t->ssn = idxsize;
    t->next = vindex;
    vindex = t;
    nopunct (t->descript);
    idxsize++;
  }
  fclose (fp);
}

void
reset_blobs ()
{
  print ("CONLOG FIPS 55-2 resource reset\r\n");
  wipe_index ();
  get_index ();
  print ("READY\r\n");
#ifdef blob_class
  print ("MYTYPE ");
  print (blob_class);
  print ("\r\n");
#endif
}

void
send_fetch_oid (struct bdex *c)
{
  char temp[namlen*2];
  sprintf (temp, "[%d %ld %lu]", c->ssn, myaddr, (unsigned long)0);
  print (temp);
}

void
do_query (char *query)
{
  struct bdex *c;
  nopunct (query);
  if (strspn (query, " ") == strlen (query))
    query[0] = '\0';
  print ("MATCHES ");

  c = vindex;
  while (c)
  {
    int tmatches, tlength;
    if (strlen (query))
    {
      if (c->descript)
      {
        tmatches = wrdmatches (c->descript, query, &tlength);
        if (tmatches == tlength)
          send_fetch_oid (c);
      }
    }
    else
    {
      /* Universal query -- limit size of response */
      if ((double)rand () / randmax < (double)universe_size/(double)idxsize)
        send_fetch_oid (c);
    }
    c = c->next;
  }
  print ("\r\n");
}

/* Respond to a FETCH from Unetd.
 */
void
fetch (char *oid)
{
  struct bdex *hit;
  char newoid[namlen*2], temp;
  int ssn;
  sscanf (oid, "[%d", &ssn);
  hit = srch_by_ssn (ssn);
  if (!hit)
    panic ("Can't find a record", 1);
  print ("0 The following datum was provided by the NIST FIPS 55-2 Geographical Database (Virginia)");
  print ("\r\n0 Thanks go to Henry Tom for this database.");
  print ("\r\n0 ID: ");
  sprintf (newoid, "%d %ld %lu", hit->ssn, myaddr, (unsigned long)0);
  print (newoid);
  print ("\r\n0 Class: ");
  print (blob_class);
  print ("\r\n0 Description: ");
  print (hit->descript);
  print ("\r\n0 FIPS 55-2 Record:  ");
  print (hit->file);
  print ("\r\n0 A brief summary of the significant features of this record follows.");
  print ("\r\n0 -     The place is called ");
  temp = hit->file[67];
  hit->file[67] = '\0';
  print (hit->file + 15);
  hit->file[67] = temp;
  print ("\r\n0 -     ");
  switch (hit->file[13]) {
  case 'C':
    print ("It's incorporated");
    break;
  case 'E':
    print ("It's an Alaska native village");
    break;
  case 'U':
    print ("It hasn't been incorporated yet");
    break;
  case 'D':
    print ("It's an Indian reservation");
    break;
  case 'H':
    print ("It's a county, or something like that");
    break;
  case 'T':
    print ("It's an Active Minor Civil Division");
    break;
  case 'Z':
    print ("It's an inactive or nonfunctioning primary county division");
    break;
  case 'A':
    print ("It's an airport");
    break;
  case 'B':
    print ("It's a post office");
    break;
  case 'G':
    print ("It's a non-government facility");
    break;
  case 'M':
    print ("It's a federal facility");
    break;
  case 'N':
    print ("It's a state, local, or international facility");
    break;
  case 'X':
    print ("Its name is obsolete or incorrect");
    break;
  default:
    print ("I can't grok the class code");
    break;
  }
  print ("\r\n0 -     It is in the state of ");
  temp = hit->file[9];
  hit->file[9] = '\0';
  print (hit->file + 7);
  hit->file[9] = temp;
  print ("\r\n1 -     Its county is ");
  temp = hit->file[92];
  hit->file[92] = '\0';
  print (hit->file + 70);
  hit->file[92] = temp;
  print ("\r\n");
}

main ()
{
#ifdef defdir
  if (chdir (defdir))
  {
    printf ("Cannot change directory to %s\n", defdir);
    puts ("Shutting down.");
    exit (-1);
  }
#endif
  if (fork ()) return 0;       /* Get away from parent. */
#ifndef nosetsid
  setsid ();                   /* Get away from controlling terminal. */
#endif
  do_sigs ();
  rand_setup ();
  create_pipe (&infd, "wres_FIPS_55-2");
  create_pipe (&outfd, "rres_FIPS_55-2");
  while (1)
    handler (get_text ());
}
