/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  rnntp.c  Turn newsgroups into a database via NNTP.
 *
 *  YOUR NNTP SERVER MUST SUPPORT XHDR AND BE QUICK ABOUT IT!
 *
 *  This is actually a heavily hacked version of rblobs.c.  Isn't
 *  everything?  =|-)
 *
 *  Defdir must contain a file called newsgroups that contains the names
 *  of the newsgroups to provide.  Don't put local newsgroups in there!
 *
 *  If you are doing a specific subtree or theme in the netnews hierarchy,
 *  change the MYTYPE to declare the right class.
 */

/* Name your NNTP server. */
#define newshost "news"

/* How often to refresh subjects.  Doing this infrequently results in a lot
 * of OOPSes.
 */
#define staleminutes 60

/* This is where the pipes go and where the newsgroups file must be. */
#define defdir "/home/faculty/alibi/bin/topdog.cs.umbc.edu"

/* As usual, I include much more than I need. */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

void
reset_blobs ();
void
fetch ();
void
print ();
void
do_query ();
void
ditch ();
char *
group2class ();
void
declare_all_classes ();
void
wipe_index ();
void
get_index ();

/* Approx. number of OID's returned for a () query */
#define universe_size 30

int infd= -1, outfd= -1, nntp= -1;
long myaddr=0, idxsize=0;
double randmax;  /* RAND_MAX + 1 */
unsigned long lasttime=0;
#define bufsize 1024
char buf[bufsize];

struct bdex {
  char *group;
  long article;
  char *subject;
  struct bdex *next;
};

struct bdex *vindex = NULL;

void
quitquitquit ()
{
  print ("BYE\r\n", outfd);
  /* Let Unetd read the BYE */
  sleep (10);
  exit (-1);
}

char *
group2class (char *fname)
{
  static char blam[bufsize];
  char *b;
  sprintf (blam, "news %s", fname);
  b = blam;
  while (*b)
  {
    if (*b == '.')
      *b = ' ';
    b++;
  }
  return blam;
}

void
timlog (char *a)
{
  print ("CONLOG ", outfd);
  print (a, outfd);
  print ("\r\n", outfd);
}

void
panic (char *why, int a)
{
  timlog (why);
  quitquitquit ();
}

#include "strfuns.c"

void
create_pipe (int *fd, char *name)
{
  *fd = -1;
  if (mknod (name, S_IFIFO | S_IREAD | S_IWRITE, NULL))
  {
    perror ("mknod");
    exit (-1);
  }
  *fd = open (name, O_RDWR, NULL);
  if (*fd == -1)
  {
    perror ("open");
    exit (-1);
  }
}

/*
 *  Line-oriented input.
 */
void
input (int s)
{
  char *a;
  a = buf;
  do {
    if (read (s, a++, 1) != 1)
    {
      if (errno != EINTR)
      {
        perror ("read");
        panic ("Read error", 1);
      }
      a--;
    }
  } while ((*(a-1) != '\n')&&(a-buf+1 < bufsize));
  *a = '\0';
#ifdef debug
  printf ("Input:  %s", buf);
  fflush (stdout);
#endif
}

void
print (char *m, int fd)
{
  write (fd, m, strlen (m));
}

/*
 *  Read an NNTP response.  Die if it's an error.
 */
void
nntp_OK ()
{
  input (nntp);
  /* Bad article number happens for cancelled posts. */
  if (!(strncmp (buf, "423", 3)))
    return;
  if (buf[0] != '2')
  {
    printf ("%s", buf);
    panic ("NNTP error", 1);
  }
}

/*
 *  Establish connection to netnews.
 */
void
NNTP_begin ()
{
  struct sockaddr_in sou;
  struct hostent *t;

  if (!(t = (struct hostent *) gethostbyname (newshost)))
  {
    perror ("gethostbyname");
    panic ("NNTP host unknown", 1);
  }
  memset ((char *) &sou, '\0', sizeof sou);
  memcpy ((char *) &sou.sin_addr, t->h_addr, t->h_length);
  sou.sin_family = (short)(t->h_addrtype);
  sou.sin_port = (u_short) (htons ((u_short)119));

  if ((nntp = socket (t->h_addrtype, SOCK_STREAM, 0)) == -1)
  {
    perror ("socket");
    panic ("Could not connect to NNTP", 1);
  }
  if (connect (nntp, &sou, sizeof sou))
  {
    perror ("connect");
    panic ("Could not connect to NNTP", 1);
  }

  nntp_OK ();
}

/*
 *  Finish up with NNTP.
 */
void
NNTP_end ()
{
  sprintf (buf, "QUIT\n");
  print (buf, nntp);
  nntp_OK ();
  close (nntp);
  nntp = -1;
}

void
check_reset ()
{
  unsigned long t;
  t = time(NULL);
  if ((t-lasttime)/60 > staleminutes)
  {
    wipe_index ();
    get_index ();
    lasttime = t;
    timlog ("Refreshed index");
  }
}

void
handler ()
{
  char tbuf[bufsize];
  strcpy (tbuf, buf);
  noeol (tbuf);
  if (strncmp (tbuf, "RESET", 5) == 0)
  {
    if (sscanf (tbuf, "RESET %ld", &myaddr) != 1)
    {
      timlog (tbuf);
      timlog ("RESET didn't provide address");
    }
    reset_blobs ();
  }
  else if (tbuf[0] == '(')
  {
    if (!myaddr)    /* Not ready yet */
    {
      timlog ("Can't answer queries until RESET provides address");
      print ("MATCHES\r\n", outfd);
    }
    else
    {
      check_reset ();
      do_query (tbuf);
    }
  }
  else if (strncmp (tbuf, "FETCH [", 7) == 0)
    fetch (tbuf+6);
  else
  {
    print ("CONLOG What is all this garbage? (see below)\r\nCONLOG ", outfd);
    print (tbuf, outfd);
    print ("\r\n", outfd);
  }
}

/* Inform Unetd that we are leaving. */
void
sig_shutdown (int sig)
{
  printf ("Received unexpected signal %d -- shutting down.\n", sig);
  quitquitquit ();
}

/* Administrator can manually send a SIGPIPE if Unetd crashes to keep from
 * having to restart the resource.
 */
void
newpipes ()
{
  signal (SIGPIPE, newpipes);
  close (infd);
  close (outfd);
  create_pipe (&infd, "wres_nntp");
  create_pipe (&outfd, "rres_nntp");
  print ("CONLOG new pipes\r\n", outfd);
}

void
do_sigs ()
{
  int a;
  for (a=1;a<NSIG;a++)
    if (a != SIGCONT)
      signal (a, sig_shutdown);
  signal (SIGPIPE, newpipes);
}

/* I got sick of compilers that didn't define RAND_MAX. */
void
rand_setup ()
{
  long temp,max=0,a;
  srand (time (NULL));
  for (a=0;a<100;a++)
  {
    temp = rand ();
    if (temp > max)
      max = temp;
  }
  if (max < 32768)
    randmax = 32768.0;
  else
    randmax = 2147483648.0;
}

void
declare_all_classes ()
{
  struct bdex *t;
  char prevclass[bufsize];
  t = vindex;
  prevclass[0] = '\0';
  while (t)
  {
    if (strcmp (t->group, prevclass))
    {
      strcpy (prevclass, t->group);
      print ("DECLASS ", outfd);
      print (group2class (prevclass), outfd);
      print ("\r\n", outfd);
    }
    t = t->next;
  }
}

void
wipe_index ()
{
  while (vindex)
  {
    struct bdex *t;
    if (vindex->subject)
      free (vindex->subject);
    if (vindex->group)
      free (vindex->group);
    t = vindex->next;
    free (vindex);
    vindex = t;
  }
  idxsize = 0;
}

struct bdex *
srch_by_grp_and_art (char *grp, long artnum)
{
  struct bdex *t = vindex;
  while (t)
  {
    if ((!(strcmp (grp, t->group)))&&(artnum==t->article))
      return t;
    t = t->next;
  }
  return NULL;
}

/* Identifiers are articlenumber:group.  Sorry for the misleading function
 * names, but like I said, it's a heavily hacked version of rblobs.c.
 */
struct bdex *
srch_by_fname (char *fname)
{
  char grp[bufsize], *a;
  long artnum;
  if (!(a = strchr (fname, ':')))
    panic ("Bogus identifier", 1);
  sscanf (a+1, "%s", grp);
  if (sscanf (fname, "%ld", &artnum) != 1)
    panic ("Bogus identifier", 1);
  return (srch_by_grp_and_art (grp, artnum));
}

void
ditch (char *grp, long art)
{
  struct bdex *p, *q;
  if ((!(strcmp (grp, vindex->group)))&&(art==vindex->article))
  {
    p = vindex;
    vindex = p->next;
    free (p->group);
    free (p->subject);
    free (p);
    idxsize--;
    return;
  }
  p = vindex;
  q = p->next;
  while (q)
  {
    if ((!(strcmp (grp, q->group)))&&(art==q->article))
    {
      p->next = q->next;
      free (q->group);
      free (q->subject);
      free (q);
      idxsize--;
      return;
    }
    p = q;
    q = q->next;
  }
  panic ("Nothing to ditch", 1);
}

void
get_index ()
{
  FILE *fp;
  long junk1, junk2, firstart, lastart;
  char grp[bufsize];
  sprintf (buf, "%s/newsgroups", defdir);
  if (!(fp = fopen (buf, "rt")))
    panic ("Can't fopen newsgroups file", 1);
  NNTP_begin ();
  while (fscanf (fp, "%s", grp) == 1)
  {
    sprintf (buf, "group %s\n", grp);
    print (buf, nntp);
    nntp_OK ();
    if ((sscanf (buf, "%ld %ld %ld %ld", &junk1, &junk2, &firstart, &lastart)) != 4)
    {
      printf ("%s", buf);
      panic ("Bogus NNTP response", 1);
    }
    sprintf (buf, "xhdr subject %ld-%ld\n", firstart, lastart);
    print (buf, nntp);
    nntp_OK ();
    input (nntp);
    while (strcmp (buf, ".\r\n"))
    {
      struct bdex *t;
      if (sscanf (buf, "%ld", &junk1) != 1)
      {
        printf ("%s", buf);
        panic ("Bogus XHDR response", 1);
      }
      noeol (buf);
      t = (struct bdex *) safmalloc (sizeof (struct bdex));
      t->group = strclone (grp);
      strcat (buf, " ");
      strcat (buf, group2class (grp));
      nopunct (buf);
      t->subject = strclone (buf);
      t->article = junk1;
      t->next = vindex;
      vindex = t;
      idxsize++;
      input (nntp);
    }

    /* Now we have to get rid of the ones that are of limited
     * distribution (what a pain)
     */
    sprintf (buf, "xhdr distribution %ld-%ld\n", firstart, lastart);
    print (buf, nntp);
    nntp_OK ();
    input (nntp);
    while (strcmp (buf, ".\r\n"))
    {
      char dist[bufsize];
      if (sscanf (buf, "%ld %s", &junk1, dist) != 2)
      {
        printf ("%s", buf);
        panic ("Bogus XHDR response", 1);
      }
      noeol (dist);
      if (strcmp (dist, "world") && strcmp (dist, "(none)"))
        ditch (grp, junk1);
      input (nntp);
    }
  }
  NNTP_end ();
  fclose (fp);
}

void
reset_blobs ()
{
  print ("CONLOG NNTP resource reset\r\n", outfd);
  check_reset ();

  /* Change the MYTYPE if you do a specific subtree or theme */
  print ("READY\r\nMYTYPE news\r\n", outfd);

  declare_all_classes ();
}

void
send_fetch_oid (struct bdex *c)
{
  /* Version numbers are not applicable. */
  sprintf (buf, "[%ld:%s %ld 0]", c->article, c->group, myaddr);
  print (buf, outfd);
}

void
do_query (char *query)
{
  struct bdex *c;
  nopunct (query);
  if (strspn (query, " ") == strlen (query))
    query[0] = '\0';
  print ("MATCHES ", outfd);

  c = vindex;
  while (c)
  {
    int tmatches, tlength;
    if (strlen (query))
    {
      if (c->subject)
      {
        tmatches = wrdmatches (c->subject, query, &tlength);
        if (tmatches == tlength)
          send_fetch_oid (c);
      }
    }
    else
    {
      /* Universal query -- limit size of response */
      if ((double)rand () / randmax < (double)universe_size/(double)idxsize)
        send_fetch_oid (c);
    }
    c = c->next;
  }
  print ("\r\n", outfd);
}

/* Respond to a FETCH from Unetd.
 */
void
fetch (char *oid)
{
  struct bdex *hit;
  char fname[bufsize];
  sscanf (oid, "[%s", fname);
  hit = srch_by_fname (fname);
  if (!hit)
  {
    /* Maybe we updated the index since Unetd got the OID. */
    print ("OOPS\r\n", outfd);
    return;
  }

  NNTP_begin ();
  sprintf (buf, "group %s\n", hit->group);
  print (buf, nntp);
  nntp_OK ();
  sprintf (buf, "article %ld\n", hit->article);
  print (buf, nntp);
  nntp_OK ();
  /* Cancelled?  Missing? */
  if (buf[0] != '2')
  {
    /* We're obviously out of date! */
    timlog ("NNTP:  missing article, updating index");
    lasttime = 0;
    print ("OOPS\r\n", outfd);
    NNTP_end ();
    return;
  }

  print ("0 The following datum was provided by an NNTP resource.", outfd);
  sprintf (buf, "\n0 ID: %ld:%s %ld 0", hit->article, hit->group, myaddr);
  print (buf, outfd);
  print ("\r\n0 Class: ", outfd);
  print (group2class (hit->group), outfd);
  print ("\r\n", outfd);

  sprintf (fname, "Description: %s\n", hit->subject);

  input (nntp);
  while (strcmp (buf, ".\r\n"))
  {
    print ("0 ", outfd);
    print (fname, outfd);
    sprintf (fname, ">%s", buf);
    input (nntp);
  }
  print ("1 ", outfd);
  print (fname, outfd);

  NNTP_end ();
}

main ()
{
#ifdef defdir
  if (chdir (defdir))
  {
    printf ("Cannot change directory to %s\n", defdir);
    puts ("Shutting down.");
    exit (-1);
  }
#endif
  if (fork ()) return 0;       /* Get away from parent. */
#ifndef nosetsid
  setsid ();                   /* Get away from controlling terminal. */
#endif
  do_sigs ();
  rand_setup ();
  create_pipe (&infd, "wres_nntp");
  create_pipe (&outfd, "rres_nntp");
  while (1)
  {
    input (infd);
    handler ();
  }
}
