/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*  rnull.c  A null resource.
 *
 *  This can be used as a starting point for building new resources or for
 *  testing purposes.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

#define defdir "/home/faculty/alibi/bin/retriever.cs.umbc.edu"

int infd, outfd;

void
print ();

void
shutdown ()
{
  print ("BYE\r\n");
  /* Let Unetd read the BYE */
  sleep (10);
  exit (-1);
}

void
create_pipe (int *fd, char *name)
{
  *fd = -1;
  if (mknod (name, S_IFIFO | S_IREAD | S_IWRITE, NULL))
  {
    perror ("mknod");
    exit (-1);
  }
  *fd = open (name, O_RDWR, NULL);
  if (*fd == -1)
  {
    perror ("open");
    exit (-1);
  }
}

void
print (char *m)
{
  write (outfd, m, strlen (m));
}

/* Get rid of trailing CR's and LF's */
char *
noeol (char *s)
{
  char *a;
  a = &(s[strlen(s)]);
  while (a > s)
    if (*(a-1) == '\n' || *(a-1) == '\r')
      *(--a) = '\0';
    else
      break;
  return s;
}

/* Get one character from the server. */
char
sgetc ()
{
  char t;
  int l=0;
  while (!l)
  {
    l = read (infd, &t, 1);
    if (l <= -1 && errno == EINTR)
    {
      l = 0;
      continue;
    }
    if (l <= -1)
    {
      perror ("read");
      puts ("Error on read from server -- shutting down.");
      shutdown ();
    }
  }
  return t;
}

/* Get a line of text */
char *
get_text ()
{
  static char buf[10000];
  char c='a';
  int l=0;
  while (c != '\n')
  {
    if (l >= 10000)
    {
      puts ("Input buffer overflow.  Shutting down.");
      shutdown ();
    }
    c = sgetc ();
    buf[l++] = c;
  }
  buf[l] = '\0';
  return buf;
}

void
handler (char *m)
{
  noeol (m);
  if (m[0] == '(')
    print ("MATCHES\r\n");
  else if (strncmp (m, "RESET", 5) == 0)
    print ("READY\r\n");
  else
  {
    print ("CONLOG What is all this garbage? (see below)\r\nCONLOG ");
    print (m);
    print ("\r\n");
  }
}

/* Inform Unetd that we are leaving. */
void
sig_shutdown (int sig)
{
  printf ("Received unexpected signal %d -- shutting down.", sig);
  shutdown ();
}

/* Administrator can manually send a SIGPIPE if Unetd crashes to keep from
 * having to restart the resource.
 */
void
newpipes ()
{
  signal (SIGPIPE, newpipes);
  close (infd);
  close (outfd);
  create_pipe (&infd, "wres_null");
  create_pipe (&outfd, "rres_null");
  print ("CONLOG new pipes\r\n");
}

void
do_sigs ()
{
  int a;
  for (a=1;a<NSIG;a++)
    if (a != SIGCONT)
      signal (a, sig_shutdown);
  signal (SIGPIPE, newpipes);
}

main ()
{
#ifdef defdir
  if (chdir (defdir))
  {
    printf ("Cannot change directory to %s\n", defdir);
    puts ("Shutting down.");
    exit (-1);
  }
#endif
  if (fork ()) return 0;       /* Get away from parent. */
#ifndef nosetsid
  setsid ();                   /* Get away from controlling terminal. */
#endif
  do_sigs ();
  create_pipe (&infd, "wres_null");
  create_pipe (&outfd, "rres_null");
  print ("CONLOG NULL RESOURCE happily up and running\r\n");
  while (1)
    handler (get_text ());
}
