/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Stuff for daemon connections */

#include "alibi.h"

/* Translate a serial number into a connection, or NULL if nonesuch. */
struct conlist *
snum2con (unsigned long snum)
{
  struct conlist *t = cons;
  while (t)
  {
    if (t->snum == snum)
      return t;
    t = t->next;
  }
  return NULL;
}

/* Detect response path lines and return 1 if they are too long. */
static int
rexpired (struct dobj *t)
{
  while (t)
  {
    char *fs;
    fs = strchr (t->text, (int)' ');
    if (fs)
    {
      fs++;
      if (!(strncmp ("Response path:", fs, 14)))
      {
        int a, b=0;
        for (a=0;a<strlen(fs);a++)
          if (fs[a] == ' ')
            b++;
        if (b < 1+TTL_resp)    /* Own name hasn't been added yet. */
          return 0;
        return 1;
      }
    }
    t = t->next;
  }
  return 0;
}

/* Deliver a datum to another daemon or to a client. */
void
forward_response (struct dobj *t, struct conlist *recipient)
{
  if (recipient->type == daemon)
  {
    /* Write updated cachedata */
    char tmp[80];
    sprintf (tmp, "0 Cachedata %lu %lu %lu\n", c_hops, c_links, c_cache);
    text_sock_write (recipient, tmp);
  }
  while (t)
  {
    char *fs;
    if (!strncmp ("0 Cachedata ", t->text, 12))
      t = t->next;     /* Cachedata MUST not be last!  (Should be first) */
    text_sock_write (recipient, t->text);
    fs = strchr (t->text, (int)' ');
    if (fs)
    {
      fs++;
      if (!(strncmp ("Response path:", fs, 14)))
      {
        /* Append own name to response path */
        text_sock_write (recipient, " ");
        text_sock_write (recipient, myhostname);
      }
    }
    text_sock_write (recipient, "\n");
    if (t->bindata)
      bin_sock_write (recipient, t->bindata, t->binlen);
    t = t->next;
  }
}

/* Some measly accounting details for delivering responses. */
void
note_rdel (struct dobj **t)
{
  int junk, mrtflag;
  struct dobj *u;
  unsigned long s_time;
  qdel++;
  {
    char *class;
    class = classify (t, NULL);
    if (class)
    {
      if (strcmp ("crap", class))
        noncrdel++;
    }
    else
      noncrdel++;
  }
  u = *t;
  mrtflag=0;
  while (u)
  {
    char junk2;
    if (sscanf (u->text, "%d Start time: %lu", &junk, &s_time) == 2)
    {
      mrt += (unsigned long)(time(NULL)) - s_time;
      mrtflag=1;
    }
    else if (sscanf (u->text, "%d This datum has been in cache since%c", &junk, &junk2) == 2)
    {
      cachedel++;
    }
    u = u->next;
  }
  if (!mrtflag)
    timlog ("No start time in response!  Mean response time will be wrong.");
}

/* Handle response messages that arrive over links. */
void
handle_response (struct conlist *t)
{
  responsecount++;
  if (strcmp (t->forsite, myhostname))
  {
    struct route_entry *rteto;
    struct conlist *giveto;

    /* Maybe cache it */
    cache_hook (&(t->datum), t->lag);

    /* Update routing info */
    qroute_hook (t);

    if (rexpired (t->datum))
    {
      char tbuf[namlen+80];
      sprintf (tbuf, "Dropping expired response for %s", t->forsite);
      timlog (tbuf);
      return;
    }

    /* Route it */
    rteto = routeto (t->forsite);
    if (!rteto)
    {
      char tbuf[namlen+80];
      sprintf (tbuf, "Dropping response for unreachable host %s", t->forsite);
      timlog (tbuf);
      return;
    }
    giveto = findcon (rteto->nexthop);
    if (!giveto)
      panic ("Could not locate nexthop in handle_response", 1);

    /* Definitely forward it */
    forward_response (t->datum, giveto);
  }
  else
  {
    struct conlist *client;
    note_rdel (&(t->datum));
    cache_hook (&(t->datum), t->lag);
    qroute_hook (t);
    client = snum2con (t->forsnum);
    if (!client)
    {
      timlog ("Dropping response for deleted or nonexistent client");
      return;
    }
    forward_response (t->datum, client);
  }
}

/* Handle response messages that we create HERE.  There are subtle semantic
 * differences which create not-so-subtle programming conflicts if you try
 * to use the same function for both.  Notice the different interface.  This
 * one also does not call the cache hook.
 *
 */
void
initiate_response (struct dobj **t, char *forsite, unsigned long forsnum)
{
  time_t ttime;
  if (strcmp (forsite, myhostname))
  {
    struct route_entry *rteto;
    struct conlist *giveto;

    /* Route it */
    rteto = routeto (forsite);
    if (!rteto)
    {
      char tbuf[namlen+80];
      sprintf (tbuf, "Dropping response for unreachable host %s", forsite);
      timlog (tbuf);
      return;
    }
    giveto = findcon (rteto->nexthop);
    if (!giveto)
      panic ("Could not locate nexthop in handle_response", 1);

    text_sock_write (giveto, "0 Cachedata 0 0 0\n");
    text_sock_write (giveto, "0 Response created at ");
    text_sock_write (giveto, myhostname);
    text_sock_write (giveto, " ");
    ttime = time((time_t *)0);
    text_sock_write (giveto, ctime(&ttime));
    forward_response (*t, giveto);
  }
  else
  {
    struct conlist *client;
    note_rdel (t);
    client = snum2con (forsnum);
    if (!client)
    {
      timlog ("Dropping response for deleted or nonexistent client");
      return;
    }
    forward_response (*t, client);
  }
}

/* Receiving a data object and some parsing */
/* The code: (bits)
 *   1 = last item in message
 *   2 = start of data block
 *   4 = question
 */
void
datumline (struct conlist *t, char *buf)
{
  struct dobj *u;
  int flag, junk1;
  char tempdest[namlen];
  unsigned long tempsnum;

  if (sscanf (buf, "%d", &flag) != 1)
    panic ("Non-datumline sent to datumline", 1);

/*
  msg = strchr (buf, (int)' ');
  if (!msg)
    panic ("Non-datumline sent to datumline", 1);
  msg++;
*/

  /* When we get the header for a block of binary data, we "sidetrack" it
   * until the block is received.  Then it comes back later along with the
   * data.
   */
  if ((flag&2)&&(!(t->bindata)))
  {
    long length;
    if (sscanf (buf, "%d Binary data follows ... %ld bytes", &junk1, &length)
    != 2)
    {
      conlog (t, "Unrecognized bindata header from");
      bozoize (t);
      return;
    }
    t->binsize = length;
    t->sidetrack = strclone (buf);
    return;
  }

  if (sscanf (buf, "%d Answer for %lu at %s", &junk1, &tempsnum, tempdest)
  == 3)
  {
    t->forsnum = tempsnum;
    strcpy (t->forsite, tempdest);
    t->dmes = answer;
  }
  else if (sscanf (buf, "%d Question from %lu at %s", &junk1, &tempsnum,
  tempdest) == 3)
  {
    t->forsnum = tempsnum;
    strcpy (t->forsite, tempdest);
    t->dmes = question;
  }

  /* Construct the datum from its pieces. */

  u = (struct dobj *) malloc (sizeof (struct dobj));
  if (!u)
    panic ("Couldn't malloc in datumline", 1);
  u->next = (struct dobj *)0;
  u->text = strclone (buf);
  u->binlen = t->binsize;
  u->bindata = t->bindata;
  t->binsize = 0;
  t->bindata = (char *)0;
  if (!(t->datum))
    t->datum = t->enddatum = u;
  else
  {
    /* Consing onto the end of a list =|-) */
    t->enddatum->next = u;
    t->enddatum = u;
  }

  if (!(flag&1))
    return;

  /* We have the complete message now */

  switch (t->dmes) {
  case question:
    handle_query (t);
    break;
  case answer:
    handle_response (t);
    break;
  default:
    conlog (t, "Garbage received from");
    bozoize (t);
  }

  free_dobj (t->datum); t->datum = t->enddatum = NULL;
  t->dmes = garbage;
}

void
daemonhandler (char *buf, struct conlist *t, struct conlist **cons)
{
  if (t->type != daemon)
    panic ("Wrong contype in daemonhandler", 1);
  if (!(strlen (buf)))
  {
    (t->blanks)++;
    if (t->blanks < 3)
      return;
    /* We've been bozo'd. */
    printf ("Bozo'd by %s.\n", t->name);
    fflush (stdout);
    delmark (t);
    return;
  }
  t->blanks = 0;
  if (strncmp (buf, "NETCHANGE", 9) == 0)
  {
    char tid[namlen], oid[namlen];
    long sd;
    sscanf (buf, "NETCHANGE %s %s %ld", tid, oid, &sd);
    if (strcmp (tid, t->name))
    {
      char buf[namlen+40];
      sprintf (buf, "Called itself %s:", tid);
      conlog (t, buf);
      bozoize (t);
    }
    else
      nc_netchange (t, oid, sd);
  }

  /* NUKEs should be ignored if this system is running production! */

  else if (strncmp (buf, "nuke", 4) == 0)
  {
#ifdef enable_nukes
    conlog (t, "Received NUKE from");
    nuke = 1;
    /* Attention:  This is a hack to avoid sending a shutdown back to */
    /* the site we got ours from.  Not that it would read it anyway. */
    delmark (t);
#else
    ;
#endif
  }
  else if (strncmp (buf, "BINDATA\r\n", 9) == 0)
  {
    /* BINDATA just flags the fact that we received some binary data.  It is
     * necessary so that the main program knows that something is going on.
     */
    ;
  }
  else if (strncmp (buf, "BIN_FINISH\r\n", 12) == 0)
  {
    /* BIN_FINISH means that the binary data object is ready to be snarfed. */
    datumline (t, t->sidetrack);
    free (t->sidetrack);
    t->sidetrack = (char *)0;
  }
  else if (buf[0] >= '0' && buf[0] <= '9')
    datumline (t, buf);
  else if (strncmp (buf, "PINGRQ", 6) == 0)
  {
    pingrqcount++;
    if (t->pingflag)
      panic ("Received PINGRQ while pingflag set", 1);
    /* Respond to pings */
    buf[5] = 'S';
    strcat (buf, "\n");
    text_sock_write (t, buf);
#ifdef lognetc
    conlog (t, "Responded to ping from");
#endif
    /* Alternate implementation of changing lag requires sending routing
     * table to t here
     */
  }
  else if (strncmp (buf, "PINGRS", 6) == 0)
  {
    long old, delay;
    char bbbuf[100];
    pingrscount++;
    if (!(t->pingflag))
      panic ("Received PINGRS while pingflag clear", 1);
#ifdef lognetc
    sprintf (bbbuf, "Received %s from", buf);
    conlog (t, bbbuf);
#endif
    sscanf (&(buf[7]), "%ld", &old);
    delay = microtime () - old;
    if (delay <= 0)
    {
      /* This might not be good in all cases, but it's important to handle */
      /* the case when microtime gets hit right where it rolls over.       */
      delay += 0x100 * 1000000;
    }
    nc_new_lag (t, max (delay, minlag));
  }
  else if (strncmp (buf, "LAG", 3) == 0)
  {
    long lag;
    lagcount++;
    if (t->pingflag)
      panic ("Received LAG while pingflag set", 1);
    sscanf (&(buf[4]), "%ld", &lag);
    nc_new_lag (t, lag);
  }
  else if (strncmp (buf, "DECLASS ", 8) == 0)
    declare_class (buf+8);
  else
  {
    conlog (t, "Running off at the socket:");
    printf ("Data received:  '%s'\n", buf);
    fflush (stdout);
    bozoize (t);
  }
}
