/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *  oids.c   Functions relating to Object ID's.
 *
 *  An Object ID consists of three fields:  datum-ID host-ID version.
 *
 *  Datum-ID is a character string (no whitespace) that uniquely identifies
 *  the datum given that the domain is restricted to the local host and version
 *  numbers are irrelevant.  For netnews, it can be the article ID; for blobs,
 *  it can be the full pathname of the file.  Host-ID is derived from the
 *  Internet address.  Version is a number used to distinguish between versions
 *  of the same datum.  It can be the numeric timestamp of the last
 *  modification of a file or just a serial number.
 *
 *  Resources.c expects OID's to be delimitable by square brackets [], so
 *  don't use them in your OID's.
 */

#include "alibi.h"

/* Return a pointer to the OID (in character string form) of a datum.  Returns
 * NULL if there is none.
 */
char *
find_oid (struct dobj *t)
{
  char foo;
  while (t)
  {
    int flag;
    if (sscanf (t->text, "%d ID:%c", &flag, &foo) == 2)
    {
      char *fs;
      fs = strchr (t->text, ':');
      fs += 2;
      return fs;
    }
    t = t->next;
  }
  return NULL;
}

int
version_blind_id_cmp (char *id1, char *id2)
{
  char *id1a, *id2a;
  int id1b, id2b;
  id1a = strrchr (id1, ' ');
  id2a = strrchr (id2, ' ');
  if ((!id1a)||(!id2a))
    return 1;
  id1b = id1a - id1;
  id2b = id2a - id2;
  if (id1b != id2b)
    return 1;
  return strncmp (id1, id2, id1b);
}
