/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*  Choose port numbers for Ubernet daemons.  A daemon must bind to one of
 *  these ports when it initiates a link to pass authentication.  The method
 *  of choosing port numbers is admittedly quite silly.
 *
 *  Port numbers above 32767 are not used.
 */

#include "alibi.h"

static int last = 13591;           /* Previously rendered port number */

/* This is a subroutine for photogenesis. */
static int
pgsub (char *buf)
{
  int r, p=0, i, digcounts[16];

  /* Blast these obsolete compilers!!! */
  /* digcounts = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}; */
  {
    int a;
    for (a=0;a<16;a++)
      digcounts[a] = 0;
  }

  /* Count up repetitions of digits */
  for (i=0;i<strlen(buf);i++)
    if (isdigit (buf[i]))
      (digcounts[(int)(buf[i])-(int)'0'])++;
    else
      (digcounts[toupper((int)(buf[i]))-(int)'A'+10])++;
  /* Repetition is photogenic. */
  for (i=1;i<16;i++)
    if (digcounts[i] > 1)
      p += digcounts[i];
  /* Zeros are bad, especially when repeated. */
  p += digcounts[0]*digcounts[0];
  /* Sequences of the same number are really bad. */
  r=0;
  for (i=1;i<strlen(buf);i++)
    if (buf[i] == buf[i-1])
      r++;
    else
    {
      p += r*r;
      r=0;
    }
  /* Ascending and descending sequences are bad. */
  for (i=1;i<strlen(buf);i++)
    if (abs ((int)(buf[i]) - (int)(buf[i-1])) == 1)
      p++;
  return p;
}

/* This function estimates how "photogenic" an integer is, i.e. how likely
 * it is that somebody else will want to use this number as the port number
 * for _their_ program.  Decimal and hexidecimal bases are considered.  There
 * are 472 completely nonphotogenic port numers.
 */
int
photogenesis (int n)
{
  int p=0;
  char buf[6];
  /* Don't give me anything over 32767 or below 10000. */
  if (n < 10000 || n > 32767)
    return 20;
  sprintf (buf, "%d", n);
  p += pgsub (buf);
  sprintf (buf, "%x", n);
  p += pgsub (buf);
  return p;
}

/* Start over. */
void
reset_port_series ()
{
  last = 13591;
}

int
next_port_number ()
{
  last++;
  while (last != 32767 && photogenesis (last))
    last++;
  if (last == 32767)
    panic ("Out of port numbers", 1);
  return last;
}
