/* -*-text-*- */

/*
    Copyright (C) 1994  David W. Flater.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*  signals.c contains signal handlers.  For the most part, these should be  */
/*  for sending commands to the local Unetd with kill.  If you want to do    */
/*  things to Unetd without taking it down, it is considerably safer to add  */
/*  signal handlers than to add capabilities to the command channel.         */

/*  Misc. notes:
 *     -- Handlers for signals which are expected more than once are re-
 *        installed each time the signal is caught since some flavors of UNIX
 *        will automatically de-install them when the signal is caught.
 *     -- Sun's old K&R compiler is said (in the man pages) to incorrectly
 *        compile code in which external variables are modified in signal
 *        handlers when the highest level of optimization is used.  Some
 *        hysterics are therefore required.
 */

#include "alibi.h"

void
sig_nuke ();
void
sig_flush ();
void
sig_pipe ();
void
sig_other ();
void
sig_reset ();
void
sig_truncate_log ();

int pipe_broke = 0, reset=0;

void
sig_setup ()
{
  int a;

  /* All unexpected signals except SIGCONT result in a shutdown. */
#ifndef sig_debug
  for (a=1;a<NSIG;a++)
    if (a != SIGCONT)
      signal (a, sig_other);
#endif

  /* The following are the expected signals. */
  signal (SIGHUP, sig_truncate_log);
  signal (SIGINT, sig_flush);
  signal (SIGPIPE, sig_pipe);
  signal (SIGUSR1, sig_nuke);
  signal (SIGUSR2, sig_reset);
}

/* Truncate stdout and stderr.  They are assumed to be truncatable.
 */
void
sig_truncate_log ()
{
  signal (SIGHUP, sig_truncate_log);
  lseek (1, 0, SEEK_SET);    /* stdout */
  lseek (2, 0, SEEK_SET);    /* stderr */
  ftruncate (1, 0);          /* stdout */
  ftruncate (2, 0);          /* stderr */
  printf ("I am %s.\n", myhostname);
  printf ("My ID number is %ld and my PID is %d.\n", myaddr, getpid ());
  timlog ("Log file(s) truncated");
}

/* Re-initialize the resources, such as to update the index. */
void
sig_reset ()
{
  signal (SIGUSR2, sig_reset);
  reset = 1;
}

void
sig_other (int sig)
{
  char buf[80];
  sprintf (buf, "Received unexpected signal %d -- shutting down.", sig);
  timlog (buf);
  exit (-1);
}

/* Flush stdout and stderr, if log files don't appear to be complete */
void
sig_flush ()
{
  signal (SIGINT, sig_flush);
  fflush (stdout);
  fflush (stderr);
}

/* This signal handler assigns to an external variable, but this code
 * should not be invoked by sane persons anyway.
 */
void
sig_nuke ()
{
  printf ("Net shutdown initiated\n");
  fflush (stdout);
  nuke = 1;
}

/* SIGPIPE is generated on write to a broken socket. */
void
sig_pipe ()
{
  signal (SIGPIPE, sig_pipe);
  pipe_broke = 1;
}

/* Indirection to accomodate old Sun compiler. */
int
broke_pipe ()
{
  return pipe_broke;
}

void
unbreak_pipe ()
{
  pipe_broke = 0;
}

int
check_reset ()
{
  int t;
  t = reset;
  reset = 0;
  return t;
}
